#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
#
# /// script
# dependencies = [
#     "daqp",
#     "meshcat",
#     "pin-pink",
#     "qpsolvers",
#     "robot_descriptions",
# ]
# ///

"""Solve IK with the Piper arm end-effector at a prescribed target."""

import sys

import meshcat_shapes
import numpy as np
import pinocchio
import pinocchio as pin
import qpsolvers
from robot_descriptions.loaders.pinocchio import load_robot_description
from scipy.spatial.transform import Rotation

import pink
from pink.tasks import FrameTask
from pink.visualization import start_meshcat_visualizer

# IK parameters
dt = 1e-2
stop_thres = 1e-8

if __name__ == "__main__":
    robot = load_robot_description("piper_description")
    model = robot.model

    viz = start_meshcat_visualizer(robot)
    viewer = viz.viewer
    meshcat_shapes.frame(viewer["end_effector_target"], opacity=0.5)
    meshcat_shapes.frame(viewer["end_effector"], opacity=1.0)

    # Frame details
    joint_name = model.names[-1]
    parent_joint = model.getJointId(joint_name)

    FRAME_NAME = "joint6"
    data = pin.Data(model)
    low = model.lowerPositionLimit
    high = model.upperPositionLimit
    q_init = pin.neutral(model)

    # Task details
    np.random.seed(0)
    q_final = np.array(
        [
            np.random.uniform(low=low[i], high=high[i], size=(1,))[0]
            for i in range(model.nq)
        ]
    )
    pin.forwardKinematics(model, data, q_final)
    target_pose = data.oMi[parent_joint]
    ee_task = FrameTask(FRAME_NAME, [1.0, 1.0, 1.0], [1.0, 1.0, 1.0])

    target = pin.SE3.Identity()
    hit_limit = "--hit-limit" in sys.argv
    target.translation = (
        np.array([0.1, 0.2, 0.3]) if hit_limit else np.array([0.0, 0.2, 0.6])
    )
    target.rotation = Rotation.from_euler("xyz", [0, 0, 0]).as_matrix()
    viewer["end_effector_target"].set_transform(target.np)
    ee_task.set_target(target)

    configuration = pink.Configuration(model, data, q_init)
    viz.display(configuration.q)
    error_norm = np.linalg.norm(ee_task.compute_error(configuration))
    print(f"Starting from {error_norm = :.2}")
    print(f"Desired precision is error_norm < {stop_thres}")

    nb_steps = 0
    while error_norm > stop_thres:
        dv = pink.solve_ik(
            configuration,
            tasks=[ee_task],
            dt=dt,
            damping=1e-8,
            solver=(
                "daqp"
                if "daqp" in qpsolvers.available_solvers
                else qpsolvers.available_solvers[0]
            ),
        )
        q_out = pin.integrate(model, configuration.q, dv * dt)
        q_out = np.clip(q_out, low, high)
        configuration = pink.Configuration(model, data, q_out)
        pin.updateFramePlacements(model, data)
        viewer["end_effector"].set_transform(
            configuration.get_transform_frame_to_world(ee_task.frame).np
        )
        viz.display(configuration.q)
        error_norm = np.linalg.norm(ee_task.compute_error(configuration))
        nb_steps += 1

    print(f"Terminated after {nb_steps} steps with {error_norm = :.2}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
#
# /// script
# dependencies = ["daqp", "meshcat", "pin-pink", "qpsolvers",
# "robot_descriptions"]
# ///

"""Standalone version of meshcat-shapes.

See <https://pypi.org/project/meshcat-shapes/>. We keep this copy in examples/
so that it can be used by examples that need it without making meshcat-shapes
(and thus meshcat) a dependency of the project.
"""

import meshcat
import numpy as np


def __attach_axes(
    handle: meshcat.Visualizer,
    length: float = 0.05,
    thickness: float = 0.002,
    opacity: float = 1.0,
) -> None:
    """Attach a set of three basis axes to a MeshCat handle.

    Args:
        handle: MeshCat handle to attach the basis axes to.
        length: Length of axis unit vectors.
        thickness: Thickness of axis unit vectors.
        opacity: Opacity of all three unit vectors.

    Note:
        As per the de-facto standard (Blender, OpenRAVE, RViz, ...), the
        x-axis is red, the y-axis is green and the z-axis is blue.
    """
    direction_names = ["x", "y", "z"]
    colors = [0xFF0000, 0x00FF00, 0x0000FF]
    rotation_axes = [[0, 0, 1], [0, 1, 0], [1, 0, 0]]
    position_cylinder_in_frame = 0.5 * length * np.eye(3)
    for i in range(3):
        dir_name = direction_names[i]
        material = meshcat.geometry.MeshLambertMaterial(
            color=colors[i], opacity=opacity
        )
        transform_cylinder_to_frame = meshcat.transformations.rotation_matrix(
            np.pi / 2, rotation_axes[i]
        )
        transform_cylinder_to_frame[0:3, 3] = position_cylinder_in_frame[i]
        cylinder = meshcat.geometry.Cylinder(length, thickness)
        handle[dir_name].set_object(cylinder, material)
        handle[dir_name].set_transform(transform_cylinder_to_frame)


def frame(
    handle: meshcat.Visualizer,
    axis_length: float = 0.1,
    axis_thickness: float = 0.005,
    opacity: float = 1.0,
    origin_color: int = 0x000000,
    origin_radius: float = 0.01,
) -> None:
    """Set MeshCat handle to a frame, represented by an origin and three axes.

    Args:
        handle: MeshCat handle to attach the frame to.
        axis_length: Length of axis unit vectors, in [m].
        axis_thickness: Thickness of axis unit vectors, in [m].
        opacity: Opacity of all three unit vectors.
        origin_color: Color of the origin sphere.
        origin_radius: Radius of the frame origin sphere, in [m].

    Note:
        As per the de-facto standard (Blender, OpenRAVE, RViz, ...), the
        x-axis is red, the y-axis is green and the z-axis is blue.
    """
    material = meshcat.geometry.MeshLambertMaterial(
        color=origin_color, opacity=opacity
    )
    sphere = meshcat.geometry.Sphere(origin_radius)
    handle.set_object(sphere, material)
    __attach_axes(handle, axis_length, axis_thickness, opacity)


def sphere(
    handle: meshcat.Visualizer,
    radius: float = 0.01,
    color: int = 0x000000,
    opacity: float = 1.0,
):
    material = meshcat.geometry.MeshLambertMaterial(
        color=color, opacity=opacity
    )
    sphere = meshcat.geometry.Sphere(
        radius,
    )
    handle.set_object(sphere, material)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

import re
import sys
from os.path import abspath, dirname, join

sys.path.insert(0, abspath(".."))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx-mathjax-offline",
    "sphinx.ext.napoleon",  # before sphinx_autodoc_typehints
    "sphinx_autodoc_typehints",
    "sphinx_favicon",
    "sphinx.ext.intersphinx",
]

# List of modules to be mocked up
autodoc_mock_imports = []

# Add any paths that contain templates here, relative to this directory.
templates_path = []

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = {'.rst': 'restructuredtext'}

# The encoding of source files.
#
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "pink"
copyright = "2022 Stéphane Caron."
author = "Stéphane Caron"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = None  # read from __init__.py

# The full version, including alpha/beta/rc tags.
release = None  # read from __init__.py

# Read version info directly from the module's __init__.py
init_path = join(dirname(dirname(str(abspath(__file__)))), "pink")
with open(f"{init_path}/__init__.py", "r") as fh:
    for line in fh:
        match = re.match(
            '__version__ = "((\\d+\\.\\d+)\\.\\d+)[a-z0-9\\-]*".*', line
        )
        if match is not None:
            release = match.group(1)
            version = match.group(2)
            break

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#
# today = ''
#
# Else, today_fmt is used as the format for a strftime call.
#
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["build", "Thumbs.db", ".DS_Store"]

# The reST default role (used for this markup: `text`) to use for all
# documents.
#
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages. See the documentation for a
# list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["css", "images"]

# These paths are either relative to html_static_path or fully qualified paths
# (eg. https://...)
html_css_files = ["custom.css"]

# Output file base name for HTML help builder.
htmlhelp_basename = "pinkdoc"

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    "papersize": "a4paper",
    # The font size ('10pt', '11pt' or '12pt').
    #
    "pointsize": "12pt",
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        project + ".tex",
        project + " Documentation",
        author,
        "howto",
    ),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#
# latex_use_parts = False

# If true, show page references after internal links.
#
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
#
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
#
# latex_appendices = []

# It false, will not define \strong, \code, 	itleref, \crossref ... but only
# \sphinxstrong, ..., \sphinxtitleref, ... To help avoid clash with user added
# packages.
#
# latex_keep_old_macro_names = True

# If false, no module index is generated.
#
# latex_domain_indices = True

favicons = [
    {
        "rel": "icon",
        "sizes": "16x16",
        "static-file": "favicon-16x16.png",
        "type": "image/png",
    },
    {
        "rel": "icon",
        "sizes": "32x32",
        "static-file": "favicon-32x32.png",
        "type": "image/png",
    },
    {
        "rel": "apple-touch-icon",
        "sizes": "180x180",
        "static-file": "apple-touch-icon-180x180.png",
        "type": "image/png",
    },
]

intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable", None),
    "qpsolvers": ("https://qpsolvers.github.io/qpsolvers/", None),
}

# Treat warnings as errors, in particular when a reference target (such as a
# :func:`something`) is not found.
nitpicky = True

nitpick_ignore = [
    ("py:class", "pinocchio.pinocchio_pywrap_default.SE3"),
    ("py:data", "typing.Optional"),
    ("py:data", "typing.Tuple"),
    ("py:data", "typing.Union"),
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
#
# /// script
# dependencies = ["clarabel", "loop-rate-limiters", "meshcat", "pin-pink",
# "qpsolvers", "robot_descriptions"]
# ///

"""Arm tracking a circular end-effector motion, use a sparse QP solver."""

import warnings

import numpy as np
import qpsolvers
from loop_rate_limiters import RateLimiter
from robot_descriptions.loaders.pinocchio import load_robot_description

import meshcat_shapes
import pink
from pink import solve_ik
from pink.tasks import FrameTask, PostureTask
from pink.visualization import start_meshcat_visualizer

# Target circle parameters
CENTER = np.array([0.3, 0.0, 0.5])  # m, in the world frame
RADIUS = 0.15  # m


if __name__ == "__main__":
    robot = load_robot_description("ur3_description", root_joint=None)

    # Initialize visualization
    viz = start_meshcat_visualizer(robot)
    viewer = viz.viewer
    meshcat_shapes.frame(viewer["target"], opacity=0.5)
    meshcat_shapes.frame(viewer["end_effector"], opacity=1.0)

    # Define differential IK tasks
    end_effector_task = FrameTask(
        "ee_link",
        position_cost=1.0,  # [cost] / [m]
        orientation_cost=0.1,  # [cost] / [rad]
    )
    posture_task = PostureTask(
        cost=1e-3,  # [cost] / [rad]
    )
    tasks = [end_effector_task, posture_task]

    # Initialize configuration
    configuration = pink.Configuration(robot.model, robot.data, robot.q0)
    for task in tasks:
        task.set_target_from_configuration(configuration)
    viz.display(configuration.q)

    # Select a sparse QP solver
    if not qpsolvers.sparse_solvers:
        raise Exception(
            "this example is for QP solvers with a sparse-matrix API, "
            "yet qpsolvers did not find any. "
            "You can install e.g. Clarabel, OSQP or SCS, then try again."
        )
    sparse_solver = qpsolvers.sparse_solvers[0]

    # Suppress sparse matrix conversion warnings
    warnings.filterwarnings(
        "ignore",
        category=qpsolvers.warnings.SparseConversionWarning,
    )

    # Control loop parameters
    rate = RateLimiter(frequency=100.0, warn=False)
    dt = rate.period
    t = 0.0  # [s]

    while True:
        # Update target to move in a circle
        target_pos = CENTER + RADIUS * np.array([np.cos(t), np.sin(t), 0.1])
        end_effector_task.transform_target_to_world.translation = target_pos

        # Update visualization frames
        viewer["target"].set_transform(
            end_effector_task.transform_target_to_world.np
        )
        viewer["end_effector"].set_transform(
            configuration.get_transform_frame_to_world(
                end_effector_task.frame
            ).np
        )

        # Solve inverse kinematics
        velocity = solve_ik(configuration, tasks, dt, solver=sparse_solver)
        configuration.integrate_inplace(velocity, dt)

        # Visualize result at a fixed framerate
        viz.display(configuration.q)
        rate.sleep()
        t += dt
